{-# LANGUAGE NoMonomorphismRestriction #-}

module Datastructures where
import Control.Exception (assert)
import Data.Sequence hiding (filter, drop)
import Data.Foldable (toList)
import Debug.Trace
import Control.Monad
import qualified Data.Map as Map
import Data.Maybe
import qualified Data.Text as Text
import Data.Char (isDigit)
import Data.List
import qualified Data.Set as Set


-- TODO + also trdy

-- Automaton:
-- The number of states, the token packet, the transitions.
-- A transition is the start- and end-state, a port-sensitive event, and a port-sensitive packet transformation.
type AutomatonS = Int
type AutomatonT packet_t = (AutomatonS,AutomatonS,Int->packet_t->Bool,Int->packet_t->Maybe (Int,packet_t))

--	Given the type of packets, there are eight primitives:
data XMAS_t packet_t = Queue Int | Source [packet_t] String | Fork | Join Int | Sink Bool | Switch (packet_t -> Bool) String | Merge | Function (packet_t -> packet_t) String
						| Automaton [String] packet_t [AutomatonT packet_t]
						| ASwitch [Maybe (packet_t -> Bool)] | Peek -- TODO remove again?
instance Show (XMAS_t packet_t) where
	show x = "{- XMAS_t object -}"
{-	Given the type of packets, a component is a primitive with an identifier.
	The ins (outs) represent the connections to initiators (targets).
	These are stored in order, beginning at port 0.
	They can be Nothing, indicating that the connection has not been determined yet.
	Otherwise, they are either an Integer, storing a pointer to the connected component or a String storing its identifier.
	Also, the port to which the component is connected is stored.
-}
data Component packet_t = Component {
	prim   :: XMAS_t packet_t,
	ide    :: String,
	ins    :: (Seq (Maybe (Either Int String,Int))),
	outs   :: (Seq (Maybe (Either Int String,Int))),
	colors :: Seq (Seq packet_t)
} deriving (Show)
--	A network is a list of components.
type Network packet_t = Seq (Component packet_t)
{-	A hierarchical network is either a flat network, or a list of networks connected by a "glue" function. 
	The glue function takes as parameters an integer $h$, a string $idx$ and a port $p$.
	Here, $h$ is the index of the current network in the hierarchy and $idx$ is the identifier of a component.
	It returns new $h'$, $idx'$ and $p'$, overwriting the old ones.
-}
data HNetwork packet_t = Flat (Network packet_t) | Hierarchy (Seq (HNetwork packet_t)) (Int -> String -> Int -> (Int,String, Int))
instance Show packet_t => Show (HNetwork packet_t) where
	show (Flat x) = "Flat " ++ show x
	show (Hierarchy xs glue) = Prelude.foldl (++) "" (map (\x -> show x ++ "\n") (toList xs))


-- A formula is either a conjunction or a disjunction of literals.
-- Literals says that a queue is full, that is contains at least one packet $d$ (at the head of the queue) or that the queue does not contain packet $d$.
-- If $d$ is Nothing, then it represents any possible color of the queue.
data Literal p_t = Is_Full Int | Is_Gt_One Int (Maybe p_t) | Is_Zero Int (Maybe p_t) | Select Int Bool | InState Int AutomatonS | NotInState Int AutomatonS | Sum_Compare [(Int,Maybe p_t)] String Int | Block Int Int (Maybe p_t) | Idle Int Int (Maybe p_t)
	deriving (Show, Eq, Ord)
data Formula p_t = AND (Set.Set (Formula p_t)) | OR (Set.Set (Formula p_t)) | T | F | Lit (Literal p_t)
	deriving (Show, Eq, Ord)

-- An SMT model assigns values to queues.
type SMTModel = Map.Map String Int


-- Apply a function of an aswitch.
appASwitchFunction :: Maybe (a -> Bool) -> a -> Bool
appASwitchFunction Nothing _  = True
appASwitchFunction (Just f) a = f a 

getStates :: [AutomatonT p_t] -> [AutomatonS]
getStates = nub . concat . map (\(s,s',e,f) -> [s,s'])



-- BUILDING AN XMAS DATASTRUCTURE

-- Find the target that is connected to out-port p of component c
find_trgt ::  Component p_t -> Int -> Network p_t -> Maybe (Either Int String,Int)
find_trgt c p x = case findIndexL (\c' -> elemIndexL (Just (Right (ide c), p)) (ins c') /= Nothing) x of
						Just indx -> let trgt = index x indx in
										Just (Right (ide trgt),fromJust $ elemIndexL (Just (Right (ide c), p)) (ins trgt))
						_ -> Nothing

-- map over all components and look for blank outs. Replace those by finding the corresponding target.
fill_blanks_outs :: Network p_t -> Network p_t
fill_blanks_outs x = mapWithIndex (\i c ->c { outs = mapWithIndex (\p si -> case si of
																	Nothing -> find_trgt c p x
																	_ -> si)
													 (outs c)} )
									x

-- Find the initiator that is connected to in-port p of component c
find_init ::  Component p_t -> Int -> Network p_t -> Maybe (Either Int String,Int)
find_init c p x = case findIndexL (\c' -> elemIndexL (Just (Right (ide c), p)) (outs c') /= Nothing) x of
						Just indx -> let init = index x indx in
										Just (Right (ide init),fromJust $ elemIndexL (Just (Right (ide c), p)) (outs init))
						_ -> Nothing

-- map over all components and look for blank ins. Replace those by finding the corresponding initiator.
fill_blanks_ins :: Network p_t -> Network p_t
fill_blanks_ins x = mapWithIndex (\i c ->c { ins =  mapWithIndex (\p si -> case si of
																	Nothing -> find_init c p x
																	_ -> si)
													  (ins c)} )
									x

-- map over all components in x and update the network structure by filling in all in- and out-channels that are Nothing
fill_blanks :: Network p_t -> Network p_t
fill_blanks = fill_blanks_ins . fill_blanks_outs


-- for sake of efficient lookups, replace the names of the component by pointers
set_pointers_c :: Int -> Component p_t -> Network p_t -> Network p_t
set_pointers_c indx c' = mapWithIndex (\n c -> c { ins  = mapWithIndex (\n i -> case i of
														 		Just (Right idx', p) -> (if idx' == ide c' then Just (Left indx,p) else i)
																_ -> i)
															 (ins c),
													 outs = mapWithIndex (\n o -> case o of
																Just (Right idx', p) -> (if idx' == ide c' then Just (Left indx,p) else o)
																_ -> o)
															 (outs c)
										})

set_pointers :: Network p_t -> Network p_t
set_pointers = join $ foldrWithIndex set_pointers_c

-- add a new component c' to the end of the network (so as to not screw up the pointers that have already been set)
add :: (XMAS_t p_t, String, [Maybe(String,Int)] , [Maybe(String,Int)]) -> Network p_t -> Network p_t
add (p,idx,ins_c,outs_c) x = (|>) x Component {prim = p, ide = idx, ins = cast_to_seq ins_c, outs = cast_to_seq outs_c, colors = fromList (Prelude.take ( Prelude.length ins_c) (repeat (fromList []))) } 
	where cast_to_seq s =  fromList (map (\si -> case si of
								Just (idx,p) -> Just (Right (idx), p)
								Nothing      -> Nothing) s)


-- build a network from a list of components (see main for examples)
build = fill_blanks . Prelude.foldr add (fromList [])

-- PRETTY PRINTING
pprint_c :: Component p_t -> String
pprint_c c = "(" ++ (ide c)
				 ++ ","
				 ++ (show (ins c))
				 ++ ","
				 ++ (show (outs c))
				 ++ ")"

pprint_n :: Network p_t -> String
pprint_n x = Prelude.foldl (++) "" (map (\c -> pprint_c c ++ "\n") (toList x))





-- FLATTENING
-- Converts a hierarchical network into a normal network.
-- Prefixes are added to the component identifiers, to avoid name clashes
-- At each of level of the hierarchy, some index $h$ in the list of subnetworks is choosen.
-- This gives us a path of integers, starting from the top level to the current.

-- A prefix is a string starting with "H", then a number, and then an underscore.
is_prefix :: String -> Bool
is_prefix s = Prelude.length s > 0 && s!!0 == 'H' && foldr (&&) True (map isDigit $ tail s)

-- Removes all prefixes
remove_prefixes :: String -> String
remove_prefixes s = intercalate "_" $ Prelude.dropWhile is_prefix $ map Text.unpack $ Text.split (=='_') $ Text.pack s

-- Makes a prefix from an Integer
mkPrefix :: Int -> String
mkPrefix h = "H" ++ show h ++ "_"

-- Makes a prefix from a path
mkPrefixPath :: [Int] -> String
mkPrefixPath pi = concat $ map mkPrefix pi

-- Finds a path to the component identified with $idx$ in the $h$th subnetwork of the hierarchy.
-- Typically, $h$ == 0 when calling this function.
find_path :: HNetwork p_t -> Int -> String -> Maybe [Int]
find_path (Flat x) h idx = case findIndexL (\c -> ide c == idx) x of
							Nothing -> Nothing
							_ 		-> Just []
find_path (Hierarchy x glue) h idx = if h >= Data.Sequence.length x then
										Nothing
									 else
										case find_path (index x h) 0 idx of
											Just pi -> Just (h:pi)
											Nothing -> find_path (Hierarchy x glue) (h + 1) idx

-- Given a path, traverse that path and then search for the component identified with $idx$.
find_path_from :: HNetwork p_t -> [Int] -> Int -> String -> Maybe [Int]
find_path_from (Flat x) 		   [] 	   h idx = assert False Nothing
find_path_from (Hierarchy xs glue) []  	   h idx = case find_path (index xs h) 0 idx of
													Nothing -> Nothing
													Just pi -> Just (h:pi)
find_path_from (Hierarchy xs glue) (h':pi) h idx = case find_path_from (index xs h') pi h idx of
													Nothing -> Nothing
													Just pi -> Just (h':pi)

-- Add prefixes to the network.
-- Takes as input the complete original hierarchical network $hx$, the glue function of the current level in the hierarchy and the current path $pi$.
-- It transforms the given flat network $x$ by adding prefixes to the ins, outs and ide's of all components, according to the path.
add_prefixes :: HNetwork p_t -> (Int -> String -> Int -> (Int,String,Int)) -> [Int] -> Network p_t -> Network p_t
add_prefixes hx glue pi x = mapWithIndex (\n c -> c { outs = mapWithIndex add_prefix_to_channel (outs c), 
													  ins  = mapWithIndex add_prefix_to_channel (ins c),
													  ide = mkPrefix (last pi) ++ ide c})
											x
	where add_prefix_to_channel p o = case o of
										Just (Right s, p) -> 
											let (h',idx',p') = glue (last pi) (remove_prefixes s) p in
												case find_path_from hx (init pi) h' idx' of
													Nothing -> Just (Right (mkPrefixPath pi ++ idx'), p')
													Just pi -> Just (Right (mkPrefixPath pi ++ idx'), p')
										_ -> o

-- Flattens the hierarchical network
-- Recursively browse through the structure. Add prefixes and then flatten it to one list. 
-- Parameter $pi$ is the current path (initially []) and the original is kept in parameter $hx$.
flatten = join $ flatten_rec []
	where {
		flatten_rec pi hx (Flat x) = x;
		flatten_rec pi hx (Hierarchy xs glue) = foldrWithIndex (\n -> (><)) (fromList [])
										$ mapWithIndex (\h -> add_prefixes hx glue (pi ++ [h])) 
										$ mapWithIndex (\h -> flatten_rec (pi ++ [h]) hx)
										xs
	}


-- FORMULAS:

disjunct :: Ord p_t => Formula p_t -> Formula p_t -> Formula p_t
disjunct T _ = T
disjunct _ T = T
disjunct F f = f
disjunct f F = f
disjunct (Lit l) (OR fs)   = OR (Set.insert (Lit l) fs)
disjunct (OR fs) (Lit l)   = OR (Set.insert (Lit l) fs)
disjunct (OR fs0) (OR fs1) = OR (Set.union fs0 fs1)
disjunct f0 f1 = if f0 == f1 then f0 else OR $ Set.fromList [f0,f1]

conjunct :: Ord p_t => Formula p_t -> Formula p_t -> Formula p_t
conjunct F _ = F
conjunct _ F = F
conjunct T f = f
conjunct f T = f
conjunct (Lit l) (AND fs)    = AND (Set.insert (Lit l) fs)
conjunct (AND fs) (Lit l)    = AND (Set.insert (Lit l) fs)
conjunct (AND fs0) (AND fs1) = AND (Set.union fs0 fs1)
conjunct f0 f1 = if f0 == f1 then f0 else AND $ Set.fromList [f0,f1]

fromBool :: Bool -> Formula p_t
fromBool True  = T
fromBool False = F


simplify_singletons :: Ord p_t => Formula p_t -> Formula p_t
simplify_singletons (OR fs) = let fs' = Set.unions $ Set.toList $ Set.map (\f -> case f of OR fs' -> fs'; _ -> Set.singleton f) (Set.map simplify_singletons fs) in
									if Set.size fs' == 1 then
										Set.findMin fs'
									else if Set.size fs' == 0 then
										F -- TODO uitzoeken waar dit vandaan komt
									else
										OR fs'
simplify_singletons (AND fs) = let fs' = Set.unions $ Set.toList $ Set.map (\f -> case f of AND fs' -> fs'; _ -> Set.singleton f) (Set.map simplify_singletons fs) in
									if Set.size fs' == 1 then
										Set.findMin fs'
									else if Set.size fs' == 0 then
										T
									else
										AND fs'

simplify_singletons f = f


